{% load i18n %}
{% load navactive %}

{% if perms.ledger.basic_access %}
<li class="nav-item">
    <a class="nav-link" href="{% url 'ledger:corporation_ledger' corporation_pk=0 %}">Corporation Ledger</a>
</li>

<li class="nav-item">
    <a class="nav-link" href="{% url 'ledger:character_ledger' character_pk=0 %}">Character Ledger</a>
</li>

<li class="nav-item">
    <a class="nav-link" href="{% url 'ledger:events_index' %}">Events</a>
</li>
{% endif %}

{% if perms.ledger.corp_audit_manager or perms.ledger.admin_access %}
<li class="nav-item">
    <span class="nav-link">|</span>
</li>
<li class="nav-item">
    <a class="nav-link" href="{% url 'ledger:ledger_add_corp' %}">Add Corporation</a>
</li>
{% endif %}

<li class="nav-item">
    <span class="nav-link">|</span>
</li>
<li class="nav-item">
    <a class="nav-link" href="{% url 'ledger:corporation_admin' %}">Corporation Overview</a>
</li>
